'''
Routines and class to support the comparison of TDS catalogs. One catalog is
assumed to be from the originating TDS data node and the other is generated by a
replicating TDS data node.

Created on Jun 21, 2010

@author: cward
@organization: USC Information Sciences Institute
'''
import logging
import sys
from lxml import etree

_events = ('start', 'end')
"""
XML parse events that trigger actions during parsing. 
@type: tuple
 """
class DatasetExtract(object):
    """
    The class defines objects that represent a subset of elements from a TDS
    catalog for a dataset.
    
    At the time this code was in development, neither a DTD nor a Schema for the
    TDS catalog XML existed. Therefore, the code may make assumptions about what
    constitutes a valid catalog file that will turn out to not be true.
    """
    __extract_properties = frozenset(['dataset_id','file_id','size','checksum',
                                      'checksum_type','tracking_id',
                                      'aggregation_id'])
    """
    These properties are the ones used to both extract and filter an 
    DatasetExtract instance. The 'aggregation_id' property can be used to 
    exclude an instance from a parent element if L{extractDataset} is called
    with C{ignore_aggregation} set to True. (This is the default value.) Only
    attributes included in the extract set may be used to compare to DataExtract
    instances.
    @note: The I{frozenset} is used as a Python approximation of the Java 
    keyword C{final}. (Although it doesn't work the same way.)
    @type: set
    """
    _checked_properties = set(['file_id','checksum','size','tracking_id'])
    """
    These properties are the ones expected to match in both instances of a 
    DatasetExtract object, if each instance has the property. This set may be
    changed at runtime.
    @type: set
    """
    _required_properties = set(['file_id','size'])
    """
    Properties that must be present for a match to be valid. This set may be
    changed at runtime.
    @type: set
    @note: If a property in the "required" set is not in the "checked" set, it
    will B{not} be checked.
    """
    @staticmethod
    def set_checked_properties(prop_set):
        """
        Override the set of properties used to compare instances of 
        DatasetExtract objects.
        @param prop_set: The new set of properties to use for comparisons.
        @type prop_set: set
        @return: Nothing
        @rtype: void
        @note: Prints a warning message to standard error if a property in the
        set is not in the L{__extract_properties} set.
        """
        DatasetExtract._checked_properties = set()
        for value in prop_set:
            if value in DatasetExtract.__extract_properties:
                DatasetExtract._checked_properties.add(value)
            else:
                print >>sys.stderr, "Value %s cannot be checked." %(value)

    @staticmethod
    def set_required_properties(prop_set):
        """
        Override the set of properties considered to be required for comparisons
        of DatasetExtract objects.
        @param prop_set: The new set of required properties.
        @type prop_set: set
        @return: Nothing
        @rtype: void
        @note: Prints a warning message to standard error if a property in the
        set is not in the L{__extract_properties} set.
        """       
        DatasetExtract._required_properties = set()
        for value in prop_set:
            if value in DatasetExtract.__extract_properties:
                DatasetExtract._required_properties.add(value)
            else:
                print >>sys.stderr, "Value %s cannot be required." %(value)

    @staticmethod
    def __compare_properties(de1, de2):
        """
        Compare the property lists of two DatasetExtract objects. The only
        properties that are checked on the ones in the L{_checked_properties}
        list. If both instances do not have a required required property
        (L{_required_properties}), the comparison returns False.
        @param de1: First object
        @type de1: DatasetExtract
        @param de2: Second object
        @type de2: DatasetExtract
        @return: Status of the comparisons.
        @rtype: boolean
        """
        for key in DatasetExtract._checked_properties:
            if key in de1.properties and key in de2.properties:
                if de1.properties[key] != de2.properties[key]:
                    print >>sys.stderr,("Property mismatch: %s not equal to %s" 
                                        %(de1.properties[key],de2.properties[key]))
                    return False
            elif key in DatasetExtract._required_properties:
                if key not in de1.properties:
                    print >>sys.stderr, ("Property %s missing from dataset %s" 
                                         %(key, de1.id))
                if key not in de2.properties:
                    print >>sys.stderr, ("Property %s missing from dataset %s"
                                         %(key, de2.id))
                return False
        return True
                
    def __init__(self):
        """
        Constructor to initialize the id, file list, and property dictionary 
        for a new instance.
        """
        self.id = None # the string ID for the dataset
        self.files = {} # dictionary of encapsulated dataset elements
        self.properties = {} # dictionary of properties for this dataset
    def __str__(self):
        """
        Returns a string of the form <datasetID:files=count:properties=count>
        where the count values are the length of the files list and properties
        dictionary.
        @rtype: string
        """
        return "<%s:files=%d:properties=%d>" % (self.id, len(self.files), 
                                                len(self.properties.keys()))
    def __repr__(self):
        """
        Returns a string of the form <datasetID|FileList|PropertyDictionary>.
        FileList is a list of nested DatasetExtract objects while 
        PropertyDictionary is a dictionary of properties for this DatasetExtract.
        @rtype: string
        """
        return "<%s|%s|%s>" % (self.id, self.files, self.properties)
    def __eq__(self,other):
        """
        This method is compares the parts of each object that need to be the
        same. See the "properties" variables for which properties are checked
        when present and which are required.
        @rtype: boolean
        """
        if isinstance(other, DatasetExtract) is not True:
            print >>sys.stderr, "Not a DatasetExtract instance."
            return False
        if self.id != other.id:
            print >>sys.stderr, "Dataset IDs do not match."
            return False
        if len(self.files) != len(other.files):
            print >>sys.stderr, "Count of files do not match."
            return False
        filelist = self.files.keys()
        for sfkey in filelist:
            if other.files.has_key(sfkey):
                sfile = self.files.get(sfkey)
                ofile = other.files.get(sfkey)
                props = DatasetExtract.__compare_properties(sfile, ofile)
                if props == False:
                    return False
            else:
                print >>sys.stderr, ("File %s not in replica catalog" %(sfkey))
                return False
        return True
        
def compareExtracts(original, replica):
    """
    Compare to dataset extracts.
    @param original: An extract from the original catalog
    @type original: L{DatasetExtract}
    @param replica: An extract from the generated replica catalog
    @type replica: L{DatasetExtract}
    @return: True if equivalent.
    @rtype: boolean
    """
    return original == replica

def createContext(cat_file):
    """
    Create the parser context for a catalog file.
    @param cat_file: Path the the catalog file to parse.
    @type cat_file: string
    @return: XML parse context
    @rtype: lxml.etree.iterparse
    """
    _trace_log.info("Enter")
    context = etree.iterparse(cat_file, events=_events)
    _trace_log.info("Return")
    return context;

def createContactFromXMLString(xml_string):
    """
    Create the parser context from an in-memory XML string representation of
    the catalog.
    @param xml_string: The XML document.
    @type xml_string: string
    @return: XML parse context
    @rtype: lxml.etree.iterparse
    """
    _trace_log.info("Enter")
    # The version of the lxml etree library we're using seems to be older. Newer
    # versions claim to be able to take the XML as a string and not require this
    # pre-processing step.
    # context = etree.iterparse(StringIO(xml_string), _events=_events)
    doc = etree.XML(xml_string)
    context = etree.iterparse(doc, events=_events)
    _trace_log.info("Return")
    return context

def extractDataset(context, ignore_aggregation=True, datasetId=None):
    """
    Iterate over the context to extract the dataset elements and relevant 
    associated data. The returned object can be used to compare this subset
    of data to another object generated from another context.
    
    @param context: A parser object.
    @type context: lxml.etree.iterparse
    @param ignore_aggregation: Flag controlling filter for aggregation elements.
    @type ignore_aggregation: boolean
    @param datasetId: ID for the top dataset element.
    @type datasetId: string
    @return: A L{DatasetExtract} object.
    @rtype: DatasetExtract
    """
    _trace_log.info("Enter")
    dataset = DatasetExtract()
    dataset.id = datasetId
    for event, elem in context:
        if event == 'start':
           if 'dataset' in elem.tag:
               _trace_log.debug("start: dataset")
               if dataset.id is None:
                   # This is the first dataset element; all others will be
                   # encapsulated in this one.
                   dataset.id = elem.get('ID')
               else:
                   next_id = elem.get('ID')
                   new_dataset = extractDataset(context,ignore_aggregation,next_id)
                   if (ignore_aggregation and 
                       ('aggregation_id' in new_dataset.properties)):
                       _trace_log.debug("excluding aggregation %s" %
                                       (new_dataset.properties['aggregation_id']))
                       continue
#                   dataset.files.append(new_dataset)
                   dataset.files[new_dataset.id] = new_dataset
           elif 'property' in elem.tag:
               name = elem.get('name')
               value = elem.get('value')
               _trace_log.debug("\tstart: property %s=%s" %(name,value))
               if (dataset.id is not None and 
                   name in DatasetExtract._DatasetExtract__extract_properties):
                   _trace_log.debug("\t-->including property %s" %(name))
                   dataset.properties[name] = value
               else:
                   _trace_log.debug("\t-->excluding property %s" %(name))
        elif event == 'end':
           if 'dataset' in elem.tag:
               _trace_log.debug("end: dataset")
               break
           elif 'property' in elem.tag:
               _trace_log.debug("\tend: property")
        else:
           pass

    _trace_log.info("returning a dataset object: %s" % dataset)
    return dataset

def _main(argv):
    """
    Test program for comparing TDS datasets.
    @param argv: Command line argument vector.
    @return: Nothing.
    @rtype: void
    """
    print "Start of run"
    orig_catalog = argv[0]
    replica_catalog = argv[1]

    context = createContext(orig_catalog)
    orig_dataset = extractDataset(context, ignore_aggregation=True, datasetId=None)
    
    context = createContext(replica_catalog)
    replica_dataset = extractDataset(context, ignore_aggregation=True, datasetId=None)

    print orig_dataset == replica_dataset
    print "End of run"
 
#------------------------------------------------------------------------------ 
# Test code, not a normal runtime entry point.
#------------------------------------------------------------------------------ 
if __name__ == '__main__':
    LOG_FILENAME="compare_test.log"
    logging.basicConfig(filename=LOG_FILENAME,
                    format='%(levelname)-10s %(funcName)s: %(message)s',
                    filemode='w',
                    level=logging.ERROR)
    _top_log   = logging.getLogger('app')
    _top_log.setLevel(logging.DEBUG)
    _trace_log = logging.getLogger('app.' + __name__)
    _main(sys.argv[1:])
else:
    _trace_log = logging.getLogger('app.' + __name__)
    """
    @type: logging.Logger
    """