"""
Driver module for creating a data transfer requests for ESG replication.
Functions in this module are used by L{esgreplicate}.

Protocols
=========
  The general requirement for replication data transfers is that the originating
  data node must provide a server that supports C{gsiftp} URIs. Any I{data 
  movement agent} that can process the gsiftp protocol can be used to download
  datasets for replication.
  
  Note that for all data movement agents, the user is expected to have already
  prepared any required security certificates or other security configurations.

Data Movement Agents
====================
  The initial release of the replication client supports two data movement
  agents, I{Bulk Data Mover}, BDM, I{globus-url-copy} and I{Globus Online}.

  BDM
  ---
    BDM is driven by a XML input file. Unless overridden by the command line 
    options, a BDM XML input file is the output from a replication download 
    request.

  globus-url-copy
  ---------------
    The gateway metadata is converted into a simple text file that has separate
    lines consisting of source/target pairs. Generate this input file using the
    C{--type globus-url-copy} switch (or C{--type guc}) on the command line for
    L{esgreplicate}.

  Globus Online Service
  ---------------------
    Globus Online is a general file transfer service capable of moving data from
    one end point to another. An input is created that may be used as input to
    the Globus Online C{transfer} command. Select this option with
    C{--type globus-online}.

Metadata as List of Dictionaries
================================
  For a simple list of files and attributes, select the option C{--type list}.
  Each line will be comprised of file_id, size, access point type, access point
  URI. This format does not setup a "target" location for files. Use this option
  only if you want to manipulate the output for use with an unsupported,
  third-party utility.

Deprecated Data Movement Agents
===============================
  In addition to the supported data movement agents, the replication client may
  be used to generate output files intended to be used with other programs or
  services. These implementations are unsupported and may require additional
  edits from the end user. Deprecated data movement agents may be moved into
  the supported category or removed entirely at a later date.

  Data Mover Lite
  ---------------
    An XML file is generated that can be submitted to DML, the Data Mover Lite,
    application. Select this option with C{--type dml}.

@author: ebrady, cward
@organization: USC Information Sciences Institute
"""
import logging
from sys import stderr
from lxml import etree as ET
from esgcet.publish.hessianlib import Hessian, ProtocolError, RemoteCallException
from replication.metaDataXML import MetaDataXML
from replication.authorizationTokens import get_authorization_tokens
from replication.tdsCatalog import get_source_catalog, store_source_catalog_map
import replication.configurationKeys as ckeys


_service = None
"""
@type: Hessian connection object
"""
_trace_log = logging.getLogger('app.' + __name__)
"""
@type: logging.Logger
"""
_global_config = None
"""
@type: ConfigParser
"""
_verbose = None
"""
@type: boolean 
"""

def create_dmr_file(parameters, config, list_of_dataset_ids):
  """
  Dispatch function for the data movement request module. Routine examines the 
  download type set in the dictionary and dispatches the call to another
  function to handle the case.
  
  @param parameters: The command line parameters.
  @type parameters: dictionary.
  @param config: Configuration data.
  @type config: ConfigParser
  @param list_of_dataset_ids: IDs for datasets to query the gateway about.
  @type list_of_dataset_ids: list.
  @return: Boolean status of results.
  @rtype: boolean
  """
  _trace_log.info("Entry: %s", parameters)
  if list_of_dataset_ids is None:
    print >> stderr, "No dataset IDs were provided."
    return False

  global _global_config
  _global_config = config
  global _verbose
  _verbose = parameters['quiet']

  status = True
  HESSIAN_URL = config.get(ckeys.SECTION, ckeys.HESSIAN_REMOTE_URL)
  PORT = config.getint(ckeys.SECTION, ckeys.HESSIAN_PORT)
  if PORT is None:
      PORT = 80
  _trace_log.debug("_service port: %d _service url: %s", PORT, HESSIAN_URL)

  # Set up hessian access, to query metadata from gateway
  global _service
  try:
      _service = Hessian(HESSIAN_URL, PORT)
  except IOError, e:
      print >>stderr, "Failed to initialize Hessian service: ", str(e)
      _trace_log.exception("Failed to initialize Hessian service:")
      return False

# Ensure that download_type has a setting, either from the command line or
# config file. Command line takes precedent.
  download_type = parameters["download_type"]
  if download_type is None:
    if config.has_option(ckeys.SECTION, ckeys.DOWNLOAD_TYPE):
      download_type = config.get(ckeys.SECTION, ckeys.DOWNLOAD_TYPE)
    else:
      download_type = ckeys.SECTION
# Same for replica_root
  if (parameters['replica_root'] is None and
      config.has_option(ckeys.SECTION, ckeys.REPLICA_ROOT)):
      parameters['replica_root'] = config.get(ckeys.SECTION, ckeys.REPLICA_ROOT)

# Use the value of download_type to select which subroutine to invoke 
  if download_type == 'bdm':
    status = _create_bdm_xml_file(parameters["output_file"],
                                  list_of_dataset_ids,
                                  int(parameters['skip_count']),
                                  parameters['replica_root'],
                                  parameters['named_access_point'])
  elif download_type == 'guc' or download_type == 'globus_url_copy':
    status = _create_guc_file_data(parameters["output_file"],
                             list_of_dataset_ids,
                             int(parameters['skip_count']),
                             parameters['replica_root'],
                             parameters['named_access_point'])
  elif download_type == 'dml':
    status = _create_dml_xml_file(parameters["output_file"],
                                  list_of_dataset_ids,
                                  parameters['named_access_point'])
  elif download_type == 'wget':
    status = _create_wget_script_file(parameters["output_file"],
                                     list_of_dataset_ids,
                                     parameters['replica_root'],
                                     parameters['named_access_point'])
  elif download_type == 'gol' or download_type == 'globus-online':
    status = _create_globusonline_transfer_file(parameters["output_file"],
                                                list_of_dataset_ids,
                                                int(parameters['skip_count']),
                                                parameters['replica_root'],
                                                parameters['named_access_point'])
  elif download_type == 'list':
      status = _create_list_file_data(parameters["output_file"],
                                     list_of_dataset_ids)
  else:
    print >> stderr, "Unknown download type: " + download_type
    status = False

  _trace_log.info("Return: status=%s", status)
  return status

def _create_bdm_xml_file(filename, list_of_dataset_ids, skip_count=0,
                         root="./", name_marker=None):
    """
    Process ESG metaData XML into an XML request file for BDM.
    @note: The generated XML file is submitted to the bdmclient program using
    the "-f" switch: C{bdmclient -f input.xml}
    @param filename: Name of a file to write request into.
    @type filename: string.
    @param list_of_dataset_ids: List of ESG datasets to request.
    @type list_of_dataset_ids: list.
    @param skip_count: Number of directories to remove from the head of the URL.
    @type skip_count: integer
    @param root: Prefix for root of the downloaded files.
    @type root: string
    @param name_marker: Name used to specify a particular protocol access point.
    @type name_marker: string
    @return: Status of results.
    @rtype: boolean
    """
    _trace_log.info("Entry: filename=%s list_of_dataset_ids=%s\nroot=%s name_marker=%s",
                   filename, list_of_dataset_ids, root, name_marker);

    datasets_found = 0
    status = False
    request = ET.Element("request")
    request.text = "\n\t"
    request.tail = "\n"

    for dataset in list_of_dataset_ids:
        # Get the meta data from the gateway
        metaData = __retrieve_metadata(dataset)
        if metaData is False:
            print >> stderr, ("Dataset " + str(dataset) + " could not be found.")
            continue
        # Get and store the original catalog for later
        if __get_source_catalog(metaData) is False:
            continue

        source_files = metaData.getLocationList("GridFTP", name_marker)
        _trace_log.debug("source_files=%s", source_files)
        source_file_count = source_files.__len__()
        if source_file_count == 0:
            print >> stderr, ("Dataset " + str(dataset) + " had no files.")
            continue

        if __gsiftp_url(source_files[0]) is not True:
            print >> stderr, ("Replication using BDM requires a gsiftp URL: " +
                              source_files[0])
            _trace_log.error("Replication using BDM requires a gsiftp URL: " +
                             source_files[0])
            continue

        datasets_found += 1 # Count as found now.
        
        # BEWARE: There's nothing that says that the meta data MUST have size
        #         checksum data for each file or that the file elements can't be
        #         mixed, i.e. some have them, then none, then some again.
        source_file_sizes = metaData.getSizeChksumList()
        source_file_size_len = source_file_sizes.__len__()
        if source_file_count != source_file_size_len:
            print >> stderr, ("WARNING: Count of source files not equal " +
                             "to count of source file sizes")
            _trace_log.warning("Count of source files not equal to count of source file sizes: %d != %d",
                              source_file_count, source_file_size_len)
        # Create a subtree for each file (if the url was provided)
        # These elements conform to the BDM XML schema
        for index, file_url in enumerate(source_files):
            if file_url is not None:
                file_request = ET.Element("file")
                file_request.text = "\n\t\t"
                source = ET.Element("source")
                source.text = file_url
                source.tail = "\n\t\t"
                file_request.append(source)

                target_path = file_url.partition("//")
                target_index = target_path[2].index("/")
                clipped_path = __clip_path(skip_count,
                                           target_path[2][target_index:])
                target_url = root + clipped_path

                target = ET.Element("target")
                target.text = target_url
                target.tail = "\n\t\t"
                file_request.append(target)

                if index < source_file_size_len :
                    size,chksum = source_file_sizes[index]
                    if size is not None:
                        target.attrib["size"] = size
                        size_request = ET.Element("size")
                        size_request.text = str(size)
                        size_request.tail = "\n\t"
                        file_request.append(size_request)
                    if chksum is not None:
                        target.attrib["checksum"] = chksum

                file_request.tail = "\n\t" # close this file and start next

                request.append(file_request)

        file_request.tail = "\n" # change tail on last element to close request

    _trace_log.info("Return: datasets_found=%d", datasets_found);

    if datasets_found > 0 and __write_xml(request, filename) is True:
        status = True
    return status

def _create_guc_file_data(filename,list_of_dataset_ids,
                          skip_count,root="./",name_marker=None):
    """
    Convert the metadata for a dataset into a file of transfer commands for
    globus-url-copy. The format of this file is "C{source target}" with one pair
    on each line.
    @param filename: Name of file to write transfer requests into.
    @type filename: string
    @param list_of_dataset_ids: List of ESG datasets to request.
    @type list_of_dataset_ids: list.
    @param skip_count: Number of directories to remove from the head of the URL.
    @type skip_count: integer
    @param root: Prefix for root of the downloaded files.
    @type root: string
    @param name_marker: Name used to specify a particular protocol access point.
    @type name_marker: string
    @return: Status of results.
    @rtype: boolean   
    """
    _trace_log.info(("Entry: filename=%s") % (filename))
    status = __globus_data_file(filename,
                                list_of_dataset_ids,
                                skip_count,
                                root,
                                name_marker)
    _trace_log.info("Return: status=%s", status)
    return status


def _create_globusonline_transfer_file(filename, 
                                       list_of_dataset_ids,
                                       skip_count,
                                       root="./",
                                       name_marker=None):
    """
    Process ESG metadata XML into an input file for the Globus Online file
    transfer service command line interface. The format of this file is
    "C{source target}" with one pair on each line. Each line may include 
    additional options for each transfer. See L{configurationKeys.GOL_OPTIONS}.
    The user is assume to have created all of the necessary accounts and
    security certificates and already knows how to use the Globus Online
    service.
    @param filename: Name of a file to write reqFalseuest into.
    @type filename: string
    @param list_of_dataset_ids: ESG datasets to request.
    @type list_of_dataset_ids: list
    @param skip_count: Number of directories to remove from the head of the URL.
    @type skip_count: integer
    @param root: Prefix for root of the downloaded files.
    @type root: string
    @param name_marker: Name used to specify a particular protocol access point.
    @type name_marker: string
    @return: Status of results.
    @rtype: boolean
    """
    _trace_log.info("Entry: filename=%s" %(filename))

    gol_options = None
        # Check if configuration environment wants to override the options.
    if _global_config.has_option(ckeys.SECTION, ckeys.GOL_OPTIONS) is True:
        gol_options = _global_config.get(ckeys.SECTION, ckeys.GOL_OPTIONS)

    status = __globus_data_file(filename,
                                list_of_dataset_ids,
                                skip_count,
                                root,
                                name_marker,
                                gol_options)
    
    _trace_log.info("Return: datasets_found=%s", status)
    return status

def _create_dml_xml_file(filename, list_of_dataset_ids, name_marker):
  """
  Process ESG metadata XML into an XML request file for DML.
  @param filename: Name of a file to write request into.
  @type filename: string
  @param list_of_dataset_ids: ESG datasets to request.
  @type list_of_dataset_ids: list
  @param name_marker: Name used to specify a particular protocol access point.
  @type name_marker: string 
  @return: Status of results.
  @rtype: boolean
  @warning: Support for DML is experimental. No schema or DTD was available at
  development time. The output file may require additional editing before use.
  """
  _trace_log.info("Entry: filename=%s list_of_dataset_ids=%s",
                 filename, list_of_dataset_ids);
  print >>stderr, "\nWARNING: DML support experimental!"
  datasets_found = 0
  status = False
### SHOULD BE ABLE TO GET THIS XML STRUCTURE FROM SOMEPLACE
  # Create base of XML request tree
  request = ET.Element("request")
  request.set("id", "dml_experimental")
  request.text = "\n  "
  request.tail = "\n"
  
  for dataset in list_of_dataset_ids:
    metaData = __retrieve_metadata(dataset)
    if metaData is False:
        print >> stderr, ("Dataset " + str(dataset) + " could not be found.")
        continue

    # Get and store the original catalog for later
    if __get_source_catalog(metaData) is False:
        continue

    # Use parsing methods to get a list of all the URLs/sizes of the files
    source_files = metaData.getLocationList("GridFTP", name_marker)
    _trace_log.debug("source_files=%s", source_files)

    if source_files is None:
      return False

    source_file_sizes = metaData.getSizeList()
    # Create a subtree for each file (if the url was provided)
    # During the development phase, no XML Schema was available for DML
    # input files. This limited testing.
    try:
      for index, file_url in enumerate(source_files):
        if file_url is not None:
          file_request = ET.Element("files")
          file_request.text = "\n    "
          url_request = ET.Element("sourceurl")
          url_request.text = file_url
          url_request.tail = "\n    "
          file_request.append(url_request)

          size = source_file_sizes[index]
          if size is not None:
            #size_request = ET.SubElement(file_request, "size")
            size_request = ET.Element("size")
            size_request.text = str(size)
            size_request.tail = "\n  "
            file_request.append(size_request)
        
        file_request.tail = "  \n  "
        request.append(file_request)
        datasets_found += 1

      file_request.tail = "\n"

    except:
      print >> stderr, ("Dataset " + str(dataset) + " could not be found.")

  _trace_log.info("Return: datasets_found=%d", datasets_found);

  if datasets_found > 0 and __write_xml(request, filename) is True:
    status = True
  return status

def _create_wget_script_file(filename,list_of_dataset_ids,
                             root="./",name_marker=None):
  """
  Process metadata XML to create a wget shell script for the download.
  @param filename: Name of a file to write request into.
  @type filename: string
  @param list_of_dataset_ids: ESG datasets to request.
  @type list_of_dataset_ids: list
  @param root: Prefix for root of the downloaded files.
  @type root: string
  @param name_marker: Name used to specify a particular protocol access point.
  @type name_marker: string
  @return: Status of results.
  @rtype: boolean
  @deprecated: B{Replication only supports GridFTP transfers.} This code cannot be
  reached.
  """
  _trace_log.info("Entry: %s, %s" %(filename,root))
  print >>stderr, "WARNING: wget support experimental!"
  status = True
  fd = None
  try:
    fd = open(filename, "w")
    fd.write("#!/bin/sh\n")
    for dataset in list_of_dataset_ids:
      metaData = __retrieve_metadata(dataset)
      if metaData is False:
          print >> stderr, ("Dataset " + str(dataset) +
                            " could not be found.")
          continue
      
        # Get and store the original catalog for later
      if __get_source_catalog(metaData) is False:
          continue

      # Use parsing methods to get a list of all the URLs/sizes of the files
      source_files = metaData.getLocationList("HTTPServer", name_marker)
      _trace_log.debug("===>source_files=%s", source_files)

      # Use the source_files list as input to routine to get authorization tokens
      # before writing to XML file.
      source_files = get_authorization_tokens(source_files, _global_config)
      if source_files is None: continue
      for source_url in source_files:
        # strip out any GET arguments
        urlpart = source_url.split('&')[0]
        parts = urlpart.split('/')
        num_parts = len(parts)
        skip = 5
        part_count = 0
        path = list()
        while part_count + skip < num_parts:
            path.append(parts[part_count + skip])
            part_count += 1
            
        filepart = '/'.join(path)
        fd.write("wget -O '%s/%s' '%s'\n" % (root, filepart, source_url))

  except EnvironmentError, e:
    print >> stderr, str(e)
    _trace_log.exception("Error writing to file %s:" % (filename))
    status = False

  except Exception, e:
    print >> stderr, ("Error processing datasets for %s: " %(filename), str(e))
    _trace_log.exception("Error processing datasets for %s:" % (filename))
    status = False

  finally:
    if fd != None:
      fd.close()

  _trace_log.info("Return: %s", status)
  return status

def  _create_list_file_data(filename,list_of_dataset_ids):
    """
    @param filename: Name of file to write file list into.
    @type filename: string
    @param list_of_dataset_ids: List of ESG datasets to request.
    @type list_of_dataset_ids: list.
    @return: Status of results.
    @rtype: boolean
    @author: Estanislao Gonzalez requested and supplied initial code for this
    function.
    """
    _trace_log.info("Entry: filename=%s list_of_dataset_ids=%s ",
                   filename, list_of_dataset_ids)

    datasets_found = 0
    status = True
    fd = None

    try:
        fd = open(filename, "w")

        for dataset in list_of_dataset_ids:
            xmlDoc = __retrieve_metadata(dataset)
            if xmlDoc is False:
                print >> stderr, ("Dataset " + str(dataset) +
                                  " could not be found.")
                continue
        # Get and store the original catalog for later
            if __get_source_catalog(xmlDoc) is False:
                continue
           
            datasets_found += 1

            metadata = xmlDoc.getMetadataList()
            if len(metadata) == 0:
                print >>stderr, "No metadata found for dataset %s." %(str(dataset))
                continue

            _trace_log.debug("===>metadata record count=%d", len(metadata))

            for dataFile in metadata:
                for ap in dataFile['access_points']:
                    fd.write("%s\t%s\t%s\t%s\n" % (dataFile["id"], 
                                                   dataFile["size"],
                                                   ap["type"], ap["uri"]))

    except EnvironmentError, e:
        print >> stderr, str(e)
        _trace_log.exception("Error writing to file %s:" % (filename))
        status = False

    except Exception, e:
        print >> stderr, "Error processing datasets for %s:" % (filename), str(e)
        _trace_log.exception("Error processing datasets for %s:" %(filename))
        status = False

    finally:
        if fd != None:
            fd.close()

    _trace_log.info("Return: datasets_found=%d", datasets_found);

    if datasets_found > 0 and status is True:
        return True
    return False

def __globus_data_file(filename,list_of_dataset_ids,skip_count,
                       root,name_marker,options=None):
    """
    Common code for writing input files for use with C{globus-url-copy} or the
    Globus Online file transfer service. For C{globus-url-copy}, just write the
    source/target pairs. For Globus Online, add any options the user has
    configured in the @C{esg.ini} file. See L{configurationKeys.GOL_OPTIONS}.
    @param filename: Name of file to write transfer requests into.
    @type filename: string
    @param list_of_dataset_ids: List of ESG datasets to request.
    @type list_of_dataset_ids: list.
    @param skip_count: Number of directories to remove from the head of the URL.
    @type skip_count: integer
    @param root: Prefix for root of the downloaded files.
    @type root: string
    @param name_marker: Name used to specify a particular protocol access point.
    @type name_marker: string
    @param options: Globus Online options to add to each transfer pair.
    @type options: string
    @return: Status of results.
    @rtype: boolean   
    """
    _trace_log.info("Entry: filename=%s list_of_dataset_ids=%s skip_count=%d " +
                    "root=%s name_marker=%s options=%s", filename,
                    list_of_dataset_ids, skip_count, root, name_marker, options)

    datasets_found = 0
    status = True
    fd = None

    try:
        fd = open(filename, "w")

        for dataset in list_of_dataset_ids:
            metaData = __retrieve_metadata(dataset)
            if metaData is False:
                print >> stderr, ("Dataset " + str(dataset) +
                                  " could not be found.")
                continue
            # Get and store the original catalog for later
            if __get_source_catalog(metaData) is False:
                continue
            # Use parsing methods to get a list of all the URLs/sizes of the files
            source_files = metaData.getLocationList("GridFTP",name_marker)
            _trace_log.debug("===>source_files=%d", len(source_files))

            if __gsiftp_url(source_files[0]) is not True:
                print >> stderr, ("Replication requires a gsiftp URL: " +
                                  source_files[0])
                _trace_log.error("Replication requires a gsiftp URL: " +
                                source_files[0])
                continue

            datasets_found += 1

            for file_url in source_files:
                target_path = file_url.partition("//")
                target_index = target_path[2].index("/")
                clipped_path = __clip_path(skip_count,
                                           target_path[2][target_index:])
                target_url = root + clipped_path
                if options is not None:
                    fd.write("\"%s\"  \"%s\" %s\n" % (file_url,
                                                      target_url,
                                                      options))
                else:
                    fd.write("\"%s\"  \"%s\"\n" % (file_url,target_url))

    except EnvironmentError, e:
        print >> stderr, str(e)
        _trace_log.exception("Error writing to file %s:" % (filename))
        status = False

    except Exception, e:
        print >> stderr, "Error processing datasets for %s:" % (filename),str(e)
        _trace_log.exception("Error processing datasets for %s:" %(filename))
        status = False

    finally:
        if fd != None:
            fd.close()

    _trace_log.info("Return: datasets_found=%d", datasets_found);

    if datasets_found > 0 and status is True:
        return True
    return False

def __clip_path(skip_count, path):
    """
    Skip over some of the directories in a path and return the remaining string.
    @param skip_count: Number of directories to clip from the front of the path.
    @type skip_count: integer
    @param path: Original path.
    @type path: string
    @rtype: string
    @raise RuntimeError: Raised if skip_count greater than count of directories
    in the path.
    """
    _trace_log.debug("Enter: %d, %s" %(skip_count, path))
    if skip_count == 0:
        clipped = path
    else:
        dirs = path.split("/")
        dircount = len(dirs) - 1
        if dircount < skip_count:
            raise RuntimeError("Can't skip %d directories when path has only %d"
                                %(skip_count, dircount)) 
        # The "skip_count" is for directories in the path,  not path separators.
        clipped = "/" + "/".join(dirs[(skip_count+1):])

    _trace_log.debug("Return: %s" %(clipped))
    return clipped

def __retrieve_metadata_xml(dataset):
    """
    Query ESG web _service for files in a dataset. Routine calls web _service to 
    retrieve metadata containing the files in a dataset.
    @param dataset: ID of one dataset.
    @type dataset: String.
    @return: The response from the gateway meta data server as a string of XML. 
    On error, returns Boolean False.
    @rtype: string or boolean
    """
    try:
        if _verbose: print "Requesting meta data for dataset %s..." % (dataset)
        _trace_log.debug("===>Looking up %s", dataset)
        # Retrieve an XML document with location information for the dataset
        xml = _service.getDatasetFiles(dataset)
        _trace_log.debug("===>Hessian Service returns %s", xml)
    except (ProtocolError, RemoteCallException), he:
        print >> stderr, ("Error communicating with web service for %s" %
                          (dataset), str(he))
        _trace_log.exception("Error communicating with web service for %s:" %
                             (dataset))
        return False
    return xml

def __retrieve_metadata(dataset):
    """
    Query ESG web _service for files in a dataset. Routine calls web _service to 
    retrieve metadata containing the files in a dataset.
    @param dataset: ID of one dataset.
    @type dataset: String.
    @return: The response from the gateway meta data server as an object
    encapsulating a XML Element object. 
    On error, returns Boolean False.
    @rtype: L{MetaDataXML} object or boolean
    """
    xmlStr = __retrieve_metadata_xml(dataset)
    if xmlStr is False:
        return False
    try:
        xmlDoc = MetaDataXML(xmlStr,dataset)
    except ET.XMLSyntaxError, e:
        print >> stderr, "XML Parse error: " + str(e)
        _trace_log.exception("XML Parse error for dataset %s:\n%s" %
                             (dataset,xmlStr))
        xmlDoc = False
    return xmlDoc
    
def __write_xml(request, filename):
    """
    Create the final XML tree, and write it to the output file.
    @param request: The the XML meta data response.
    @type request: String.
    @param filename: Name of output file.
    @type filename: String.
    @return: Boolean indicating status.
    @rtype: boolean
    """
    tree = ET.ElementTree(request)
    try:
        tree.write(filename,encoding="UTF-8",xml_declaration=True,method="xml")
    except EnvironmentError, inst:
        print >> stderr, str(inst)
        _trace_log.exception("Unexpected error writing to file %s:"%(filename))
        return False

    return True

def __get_source_catalog(metaData):
    """
    Extract the URI of the original TDS catalog, retrieve and store it locally.
    If the meta data does not contain the URI of the source catalog, this
    function returns false and processing will continue with the next dataset.
    @param metaData: The meta data XML encapsulated in a L{MetaDataXML} object.
    @type metaData: MetaDataXML
    @return: Boolean status, True on success, else False
    """
    source_catalog_uri = metaData.getThreddsCatalogUri()
    _trace_log.debug('source_catalog_uri=%s', source_catalog_uri)
    if source_catalog_uri is None or len(source_catalog_uri) == 0 :
        print >> stderr, "No value in source_catalog_uri attribute"
        return False
    try:
        __store_source_catalog(metaData.getDatasetId(), source_catalog_uri)
    except EnvironmentError, e:
        print >> stderr, str(e)
        _trace_log.exception("Environment error storing TDS catalog")
        return False
    return True
    

def __store_source_catalog(dataset, source_catalog_uri):
    """
    Store the source data node TDS catalog for later use by publication.
    @param dataset: ID of the dataset associated with the catalog.
    @type dataset: string
    @param source_catalog_uri: URI of the catalog.
    @type source_catalog_uri: string
    @return: Nothing.
    @rtype: void
    @raise IOError: An IOError exception may be thrown when writing the file.
    @raise HTTPError: An HTTPError exception may be thrown when downloading file.
    @raise RuntimeError: If a value for the temporary TDS directory is not found.
    """
    if _global_config.has_option(ckeys.SECTION, ckeys.TDS_TEMP_DIR_KEY):
        tds_temp_dir = _global_config.get(ckeys.SECTION, ckeys.TDS_TEMP_DIR_KEY)
    else:
        raise RuntimeError("No value found in ini file for %s"
                           %(ckeys.TDS_TEMP_DIR_KEY))
        
    catalog_name = get_source_catalog(source_catalog_uri, tds_temp_dir)
    store_source_catalog_map(dataset, catalog_name, tds_temp_dir)
    return

def __gsiftp_url(url):
    """
    Check the URL to see that it uses gsiftp.
    @param url: URL to check.
    @type url: string
    @return: True if it is a gsiftp URL, else False.
    @rtype: boolean
    """
    is_gsiftp = url.startswith("gsiftp")
    return is_gsiftp

