'''
Class to re-implement the functions in parseMetadata.py. This reduces the number
of times the XML string gets converted into a lxml.etree._Element object.

Created on Jun 3, 2011

@author: Craig Ward
@author: Erin Brady
@organization: USC Information Sciences Institute
'''
import logging
import sys
from lxml import etree

class MetaDataXML(object):
    '''
    Wrapper class around an etree Element to facilitate extraction and
    processing of ESG Gateway meta data.
    '''
    def __init__(self,xmlString,datasetId=None):
        '''
        Constructor uses the XML document contained in the string to create an
        lxml.etree.Element to represent the document internally. If the
        datasetId is not provided, it will be assigned the Python "None" value.
        @param xmlString: The XML document as a string.
        @type xmlString: string
        @param datasetId: Dataset ID of the ESG dataset being processed.
        @type datasetId: string
        @raise lxml.etree.XMLSyntaxError: Raised if the string does not
        represent a valid XML document.
        '''
        self.docRoot = etree.fromstring(xmlString, parser=None, base_url=None)
        self.esg_namespace = self.docRoot.nsmap.get("esg")
        self.esg_dataset = datasetId

    def getDatasetId(self):
        """
        Return the dataset ID for this meta data object. The value is either set
        at instance creation or it is looked up in the "id" attribute of the top
        level dataset element.
        @return: Dataset ID
        @rtype: string
        """
        if self.esg_dataset is not None:
            return self.esg_dataset
        dataset = self.docRoot.xpath('/t:ESG/t:dataset',
                                     namespaces={'t': self.esg_namespace})
        self.esg_dataset = dataset[0].get("id")
        return self.esg_dataset
    
    def getLocationList(self, protocol="GridFTP", name_marker="GridFTP-BDM"):
        """
        Extract the files in the data set. The returned list includes the URL
        protocol. The meta data may have multiple access protocols defined for
        a file. Use the value passed as protocol as selector.
        @param protocol: Server protocol to select for. Default is "GridFTP."
        @type protocol: string
        @param name_marker: Expected name of the service when more-than-one service
        uses the same protocol. This is how "vanilla" GridFTP services are 
        distinguished from ESG-specific (BDM-enabled) GridFTP services. Set to
        "None" to have the value ignored.
        @type name_marker: string
        @return: A list of the URLs for the files in a data set.
        @rtype: list 
        """
        _trace_log.info("Entry: Seeking %s data_access_capability by name %s" % 
                        (protocol, name_marker))
        list_of_files = []

        base_uri_element = self.docRoot.xpath('/t:ESG/t:dataset/t:data_access_capability',
                                              namespaces={'t': self.esg_namespace})
        base_uri = None
        for element in base_uri_element:
            protocol_type = element.get('type')
            _trace_log.debug("Comparing %s to %s" % (protocol_type, protocol))
            if protocol_type.lower() != protocol.lower():
                continue
            capability_name = element.get('name')
            base_uri = element.get('base_uri')
        # If the caller hasn't requested a name marker, we're done.
            if name_marker == None:
                break
        # See if this element has the name the caller is looking for.
            if capability_name.lower() != name_marker.lower():
                base_uri = None # reset this "flag"
                continue
            break

        if base_uri is None:
            if name_marker is None:
                print >> sys.stderr, "No data_access_capability of protocol %s" % (protocol)
                _trace_log.error("Error Return: No data_access_capability of protocol %s" % (protocol))
            else:
                print >> sys.stderr, "No data_access_capability of protocol %s marked by %s" % (protocol, name_marker)
                _trace_log.error("Error Return: No data_access_capability of protocol %s marked by %s" % (protocol, name_marker))
            return list_of_files # An empty list

        if base_uri.endswith('/'):
            pass
        else:
            base_uri = base_uri + '/'

        files_to_transfer = self.docRoot.xpath('/t:ESG/t:dataset/t:files/t:file/t:file_access_point', 
                                               namespaces={'t': self.esg_namespace})
        for element in files_to_transfer:
            capability = element.get('data_access_capability')
            name = element.get("uri")
            if name.startswith('/'):
                name = name.lstrip('/')
            _trace_log.debug("Checking capability %s" % (capability_name))
            if capability == capability_name:
                _trace_log.debug("Adding %s to list" % (name))
                list_of_files.append(base_uri + name)

        _trace_log.info("Normal Return")
        return list_of_files # List with actual files.
    
    def getSizeList(self):
        """
        Extract file sizes from ESG Gateway XML meta data. If the size is not
        present, it is added as the Python "None" value.
        @return: List of sizes for files.
        @rtype: list
        """
        list_of_sizes = []

        file_elements = self.docRoot.xpath('/t:ESG/t:dataset/t:files/t:file', 
                                           namespaces={'t': self.esg_namespace})
        for element in file_elements:
            size = element.get("size")
            list_of_sizes.append(size)

        return list_of_sizes

    def getSizeChksumList(self, algorithm="md5"):
        """
        Extract file sizes and checksums from ESG Gateway XML meta data. If
        either or both are not present, they are added to each tuple as the
        Python "None" value.
        @param algorithm: The checksum algorithm to select when picking checksum
        elements to include in the tuple.
        @type algorithm: string
        @return: List of tuples with sizes and checksums (in that order) for files.
        @rtype: list
        """
        list_of_attributes = []

        file_elements = self.docRoot.xpath('/t:ESG/t:dataset/t:files/t:file',
                                           namespaces={'t': self.esg_namespace})
        for element in file_elements:
            checksum_value = None
            for checksum in element:
                # This will pickup more than just checksum elements, but that
                # doesn't seem to require the additional complexity to filter.
                algo = checksum.get("algorithm")
                if algo is not None and algorithm.lower() == algo.lower():
                    checksum_value = checksum.get("value")
                    break
            size = element.get("size")
            list_of_attributes.append((size,checksum_value))

        return list_of_attributes

    def getThreddsCatalogUri(self):
        """
        Retrieve the URL of the source data node TDS catalog from the results of a
        meta data query to an ESG gateway node.
        @return: The URL to the source data node TDS catalog as a string.
        @rtype: string
        """
        dataset = self.docRoot.xpath("/t:ESG/t:dataset",
                                     namespaces={'t': self.esg_namespace})
        #attributes = dataset.attrib
        #source_uri = attributes["source_catalog_uri"]
        source_uri = dataset[0].get("source_catalog_uri")
        return source_uri

    def getMetadataList(self):
        """
        Convert the metadata into a Python list of dictionary objects. The XML is
        passed in as a string. An XML document object is created internally. Uses
        XPath queries. The returned data structure has the format
        C{[file_id, size, checksum, checksum_type, [access_points]]}. Other metadata
        ignored.
        @return: Metadata stored as a list of Dictionary objects.
        @rtype: list
        @author: Estanislao Gonzalez requested and supplied the initial code for
        this function.
        """
        list_of_dictionaries = []
    
        file_elements = self.docRoot.xpath('/t:ESG/t:dataset/t:files/t:file',
                                               namespaces={'t': self.esg_namespace})
        for element in file_elements:
            name = element.get("name")
            size = element.get("size")
            #fap = file access point
            faps = []
            for fap in element:
                faps.append({"uri": fap.get("uri"),
                             "type": fap.get("data_access_capability")})            
        
            checksum = element.get("checksum")
            checksum_type = element.get("checksum_type")
            list_of_dictionaries.append({ "id":name, "size":size,
                                          "checksum":checksum,
                                          "checksum_type":checksum_type,
                                          "access_points":faps})
    
        return list_of_dictionaries

    def __repr__(self):
        """
        Serialize the XML element as a string. This string could be used to
        recreate the instance.
        @return: XML string representing the document.
        @rtype: string
        """
        return etree.tostring(self.docRoot)

    def __str__(self):
        """
        Format a string representation of the instance illustrating the active
        properties. This does not display the original XML document.
        @return: String representation of the instance.
        @rtype: string
        """
        
        if self.esg_dataset is None:
            return "<None:" + self.esg_namespace + ":" + str(self.docRoot) + ">"
        return ("<" + self.esg_dataset + ":" + self.esg_namespace + ":" + 
                self.docRoot.__str__() + ">")


def __main(args):
    """
    This function is for testing the MetaDataXML class; it is not an entry point
    into the program. The command line argument is the name of a file to open 
    and attempt to process as an XML meta data document.
    @param args: Command line with the name of the test XML file.
    @type args: list
    @rtype: integer
    @return: 0 always.
    """
    try:
        # Test by passing a name of a file to process.
        testXmlFile = open(args[0])
        testXml = testXmlFile.read()
        testXmlFile.close()
        metaData = MetaDataXML(testXml)
    except EnvironmentError, e:
        print "EnvironmentError: " + e
        return 0
    except etree.XMLSyntaxError, e:
        print "Parse error: " + str(e)
        return 0

    print "Testing __str__"
    print metaData

    print "Testing getDatasetId"
    retval = metaData.getDatasetId()
    print retval
    print "Testing __str__ again"
    print str(metaData)

    print "Testing repr()"
    print repr(metaData)

    print "Testing getLocationsList"
    retval = metaData.getLocationList("GridFTP", "GridFTP-BDM")
    print retval
    
    print "Testing getSizeList"
    retval = metaData.getSizeList()
    print retval
    
    print "Testing getSizeChksumList"
    retval = metaData.getSizeChksumList()
    print retval
    
    print "Testing getThreddsCatalogUri"
    retval = metaData.getThreddsCatalogUri()
    print retval
    
    print "Testing getMetadataList"
    retval = metaData.getMetadataList()
    print retval
    
    return 0

#------------------------------------------------------------------------------ 
# Test code, not a normal runtime entry point.
#------------------------------------------------------------------------------ 
if __name__ == '__main__':
    LOG_FILENAME="metadata_xml_test.log"
    logging.basicConfig(filename=LOG_FILENAME,
                    format='%(levelname)-10s %(funcName)s: %(message)s',
                    filemode='w',
                    level=logging.ERROR)
    _top_log   = logging.getLogger('app')
    _top_log.setLevel(logging.DEBUG)
    _trace_log = logging.getLogger('app.' + __name__)
    __main(sys.argv[1:]) # strip python script name
else:
    _trace_log = logging.getLogger('app.' + __name__)
    """
    @type: logging.Logger
    """
