'''
Routines for managing the Thredds catalogs from source data nodes. These files
are used as part of the sanity checks for replicated ESG data.

Created on May 25, 2010

@author: cward
@organization: USC Information Sciences Institute
'''
#import sys
import logging
try:
    from urllib.request import urlopen, Request, URLError, HTTPError  # Python 3
except ImportError:
    from urllib2 import urlopen, Request, URLError, HTTPError        # Python 2

_trace_log = logging.getLogger('app.' + __name__)
"""
@type: logging.Logger
"""
mapdb_file = 'catalog_map.txt'
"""
Name of file with mappings of dataset IDs to source TDS catalog XML files.
@type: string
"""

def get_source_catalog(url, dir_name="./"):
    """
    Download a source Thredds Catalog for later use in publication process.
    Returns path to the downloaded file as a string.
    @param url: URL of the catalog
    @type url: string
    @param dir_name: Directory location, with closing delimiter, to place catalog file.
    @type dir_name: string
    @return: Path to the stored catalog file.
    @rtype: string
    @raise URLError: On error with underlying network problems (e.g. host does not respond)
    @raise HTTPError: On an HTTP protocol error.
    @raise EnvironmentError: On IO Error writing  file.
    """
    _trace_log.info("Entry: %s into %s", url, dir_name)
    try:
        req = Request(url)
        connection = urlopen(req)
        catalog = connection.read();
        connection.close();

        filepart = req.get_selector().rindex("/")
        if dir_name.endswith("/"):
            destination = dir_name + req.get_selector()[(filepart+ 1):]
        else:
            destination = dir_name + req.get_selector()[(filepart):]

        _trace_log.info("\tSaving to file %s", destination)
        outfile = open(destination, 'w')
        print >> outfile, catalog
        outfile.close();
    except HTTPError:    # use "as e" for Python 3
        _trace_log.exception("HTTPError exception requesting %s:" %(url))
        raise
    except URLError:
        _trace_log.exception("URLError exception getting URL %s:" %(url))
        raise
    except EnvironmentError: # use "as e" for Python 3
        if destination is not None:
            _trace_log.exception("EnvironmentError saving %s:" % (destination))
        else:
            _trace_log.exception("EnvironmentError processing %s:" % (url))
        raise
    
    _trace_log.info("Returning \"%s\"", destination)
    return destination

#------------------------------------------------------------------------------
# store_source_catalog_map
#------------------------------------------------------------------------------
def store_source_catalog_map(dataset_id, catalog, db_dir_name="./"):
    """
    Routine to record which dataset id goes with which TDS catalog file. Operation
    is an append.
    @return: No return value.
    @rtype: void
    @param dataset_id: String representing id of the dataset.
    @type dataset_id: string  
    @param catalog: Name of the catalog file.
    @type catalog: string    
    @param db_dir_name: Path to of directory to find/create the file.
    @type db_dir_name: string
    @raise IOError: on file I/O error.    
    """
    _trace_log.info("Entry: %s, %s, %s", dataset_id, catalog, db_dir_name)

    try:
        dbfile = _construct_mapdb_filename(db_dir_name)
        db = open(dbfile,"a")
        print >>db, "%s : %s" % (dataset_id, catalog)
        db.close();
    except IOError:
        _trace_log.exception("IOError")
        raise

    _trace_log.info("Return")
    return

#------------------------------------------------------------------------------
# get_srocue_catalog_for_dataset
#------------------------------------------------------------------------------
def get_source_catalog_for_dataset(dataset_id, db_dir_name):
    """
    Retrieve the path to a source TDS catalog for a dataset by ID.
    @param dataset_id: ID for the dataset to lookup.
    @type dataset_id: string
    @param db_dir_name: Path to directory with originating TDS catalog files.
    @type db_dir_name: string
    @return: Path to the originating TCS catalog file for the dataset.
    @rtype: string
    @raise EnvironmentError: Traps, logs, and raises again.
    """
    _trace_log.info("Entry: %s, %s" % (dataset_id, db_dir_name))
    source_catalog = None
    try:
        dbfile = _construct_mapdb_filename(db_dir_name)
        db = open(dbfile,"r")
        line = db.readline()
        while line:
            dset, source_catalog = line.split(':')
            dset = dset.strip()
            if dset == dataset_id:
                break
            line = db.readline()
        source_catalog = source_catalog.strip()
        db.close()
    except EnvironmentError:
        source_catalog = None
        _trace_log.exception("Error getting source catalog:")
        raise
    
    _trace_log.info("Return: %s" %(source_catalog))
    return source_catalog

#------------------------------------------------------------------------------
# _construct_mapdb_filename
#------------------------------------------------------------------------------ 
def _construct_mapdb_filename(dirpath):
    """
    Append the name used for the source catalog mapping database to the
    directory path specified.
    @param dirpath: Path to directory with the mapping file.
    @type dirpath: string
    @return: Path to the mapping file.
    @rtype: string
    """
    if dirpath.endswith('/'):
        dbpath = dirpath + mapdb_file
    else:
        dbpath = dirpath + "/" + mapdb_file
    _trace_log.debug("constructed dbpath = %s", dbpath)
    return dbpath
